# -*- coding: utf-8 -*-


''' Gets the list of the packages that can be downloaded.
'''

import urllib
import urlparse
from collections import namedtuple
from os import listdir
from os.path import join, exists, basename

from bs4 import BeautifulSoup
from flask import abort, render_template
from requests import get

from flask_pypi_proxy.app import app
from flask_pypi_proxy.utils import (get_package_path, get_base_path,
                                    is_private, url_is_egg_file)


VersionData = namedtuple('VersionData', ['name', 'md5', 'external_link'])


@app.route('/simple/')
def simple():
    ''' Return the template which list all the packages that are installed
    '''
    packages = []
    for filename in listdir(get_base_path()):
        packages.append(filename)
    return render_template('simple.html', packages=packages)


@app.route('/simple/<package_name>/')
def simple_package(package_name):
    ''' Given a package name, returns all the versions for downloading
    that package.

    If the package doesn't exists, then it will call PyPi (CheeseShop).
    But if the package exists in the local path, then it will get all
    the versions for the local package.

    This will take into account if the egg is private or if it is a normal
    egg that was uploaded to PyPi. This is important to take into account
    the version of the eggs. For example, a proyect requires request==1.0.4
    and another package uses request==1.0.3. Then the instalation of the
    second package will fail because it wasn't downloaded an the **request**
    folder only has the 1.0.4 version.

    To solve this problem, the system uses 2 different kinds of eggs:

    * private eggs: are the eggs that you uploaded to the private repo.
    * normal eggs: are the eggs that are downloaded from pypi.

    So the normal eggs will always get the simple page from the pypi repo,
    will the private eggs will always be read from the filesystem.


    :param package_name: the name of the egg package. This is only the
                          name of the package with the version or anything
                          else.

    :return: a template with all the links to download the packages.
    '''
    app.logger.debug('Requesting index for: %s', package_name)
    package_folder = get_package_path(package_name)
    if (is_private(package_name) or (
            exists(package_name) and app.config['SHOULD_USE_EXISTING'])):

        app.logger.debug('Found information of package: %s in local repository',
                         package_name)
        package_versions = []
        template_data = dict(
            source_letter=package_name[0],
            package_name=package_name,
            versions=package_versions
        )

        for filename in listdir(package_folder):
            if not filename.endswith('.md5'):
                # I only read .md5 files so I skip this egg (or tar,
                # or zip) file
                continue

            with open(join(package_folder, filename)) as md5_file:
                md5 = md5_file.read(-1)

            # remove .md5 extension
            name = filename[:-4]
            data = VersionData(name, md5, None)
            package_versions.append(data)

        return render_template('simple_package.html', **template_data)
    else:
        app.logger.debug('Didnt found package: %s in local repository. '
                         'Using proxy.', package_name)
        url = app.config['PYPI_URL'] + 'simple/%s/' % package_name
        response = get(url)

        if response.status_code != 200:
            app.logger.warning('Error while getting proxy info for: %s'
                               'Errors details: %s', package_name,
                               response.text)
            abort(response.status_code)

        if response.history:
            app.logger.debug('The url was redirected')
            # in this case, the request was redirect, so I should also
            # take into account this change. For example, this happens
            # when requesting flask-bcrypt and on Pypi the request is
            # redirected to Flask-Bcrypt
            package_name = urlparse.urlparse(response.url).path
            package_name = package_name.replace('/simple/', '')
            package_name = package_name.replace('/', '')

        content = response.content
        external_links = set()

        # contains the list of pacges whih where checked because
        # on the link they had the information of
        visited_download_pages = set()
        soup = BeautifulSoup(content)
        package_versions = []

        for panchor in soup.find_all('a'):
            if panchor.get('rel') and panchor.get('rel')[0] == 'homepage':
                # skip getting information on the project homepage
                continue

            href = panchor.get('href')
            app.logger.debug('Found the link: %s', panchor.get('href'))
            if href.startswith('../../packages/'):
                # then the package is hosted on pypi.
                pk_name = basename(href)
                pk_name, md5_data = pk_name.split('#md5=')
                pk_name = pk_name.replace('#md5=', '')
                data = VersionData(pk_name, md5_data, None)
                package_versions.append(data)
                continue

            parsed = urlparse.urlparse(href)
            if parsed.hostname:
                # then the package had a full path to the file
                if parsed.hostname == 'pypi.python.org':
                    # then it is hosted on the pypi server, so I change
                    # it to make it a relative url
                    pk_name = basename(parsed.path)
                    if '#md5=' in parsed.path:
                        pk_name, md5_data = pk_name.split('#md5=')
                        pk_name = pk_name.replace('#md5=', '')
                    else:
                        md5_data = ''
                    data = VersionData(pk_name, md5_data, None)
                    package_versions.append(data)

                else:
                    # the python package is hosted on another server
                    # that isn't pypi. The packages that doesn't have
                    # rel=download, then they are links to some pages
                    if panchor.get('rel') and panchor.get('rel')[0] == 'download':
                        if url_is_egg_file(parsed.path):
                            external_links.add(href)
                        else:
                            # href point to an external page where the links
                            # to download the package will be found
                            if href not in visited_download_pages:
                                visited_download_pages.add(href)
                                external_links.update(find_external_links(href))

        # after collecting all external links, we insert them in the html page
        for external_url in external_links:
            package_version = basename(external_url)
            existing_value = filter(lambda pv: pv.name == package_version,
                                    package_versions)
            external_link = urllib.urlencode({'remote': external_url})
            if existing_value:
                package_versions.remove(existing_value[0])

            # check that the package version doens't override the one that
            # already exists on pypi
            existing_data = filter(lambda v: v.name == package_version,
                                   package_versions)
            if existing_data:
                continue

            data = VersionData(package_version, '', external_link)
            package_versions.append(data)

        template_data = dict(
            source_letter=package_name[0],
            package_name=package_name,
            versions=package_versions
        )
        return render_template('simple_package.html', **template_data)


def find_external_links(url):
    '''Look for links to files in a web page and returns a set.
    '''
    links = set()
    try:
        response = get(url)
        if response.status_code != 200:
            app.logger.warning('Error while getting proxy info for: %s'
                               'Errors details: %s', url,
                               response.text)
        else:
            content_type = response.headers.get('content-type', '')
            if content_type in ('application/x-gzip'):
                # in this case the URL was a redirection to download
                # a package. For example, sourceforge.
                links.add(response.url)
                return links
            if response.content:
                soup = BeautifulSoup(response.content)
                for anchor in soup.find_all('a'):
                    href = anchor.get("href")
                    if url_is_egg_file(href):
                        # href points to a filename
                        if not url.endswith('/'):
                            url += '/'
                        href = get_absolute_url(href, url)
                        links.add(href)
    except:
        # something happened when looking for external links:
        #       timeout, HTML parser error, etc.
        # we must not fail and only log the error
        app.logger.exception('')
    return links


def get_absolute_url(url, root_url):
    '''Make relative URLs absolute

    >>> get_absolute_url('/src/blah.zip', 'https://awesome.org/')
    'https://awesome.org/src/blah.zip'
    >>> get_absolute_url('http://foo.bar.org/blah.zip', 'https://awesome.org/')
    'http://foo.bar.org/blah.zip'
    '''
    parsed = urlparse.urlparse(url)
    if url.startswith('//'):
        # this are the URLS parsed from code.google.com
        return 'http:' + url
    elif parsed.scheme:
        return url
    else:
        return urlparse.urljoin(root_url, parsed.path)



# -*- coding: utf-8 -*-

from flask_pypi_proxy.app import app
from os.path import join
from hashlib import md5


def is_private(egg_name):
    ''' Checks if the egg_name is private or if belongs to one of
    the eggs that are uploaded to the normal pypi.

    :param str egg_name: the name of the egg without the version information.

    :return: true if the egg is private.
    '''
    return egg_name in app.config['PRIVATE_EGGS']


def get_base_path():
    ''' Gets the base path where all the eggs are on this servers
    '''
    return app.config['BASE_FOLDER_PATH']


def get_package_path(egg_name):
    ''' Given the name of a package, it gets the local path for it.

    :param egg_name: the name (it might also include the version) of
                     a python package.

    :return: the local path where the file can be found on the local
             system.
    '''
    return join(get_base_path(), egg_name)


def get_md5_for_content(package_content):
    ''' Given the content of a package it returns the md5 of the file.
    '''
    res = md5(package_content)
    return res.hexdigest()


def url_is_egg_file(url):
    return url is not None and (   url.lower().endswith('.zip')
                                or url.lower().endswith('.tar.gz')
                                or url.lower().endswith('.egg')
                                or url.lower().endswith('.exe')
                                or url.lower().endswith('.msi'))



# -*- coding: utf-8 -*-

''' Handles downloading a package file.

This is called by easy_install or pip after calling the simple, and getting
the version of the package to download.

'''

import magic
from flask import make_response, request, abort
from flask_pypi_proxy.app import app
from flask_pypi_proxy.utils import (get_base_path, get_package_path,
                                    get_md5_for_content)
from os import makedirs
from os.path import join, exists
from requests import get, head
# from subprocess import Popen


# robin-jarry: changed the "source" to a variable placeholder in the URL to
# deal with non-source packages (usually stored in /packages/<python version>/...)
@app.route('/packages/<package_type>/<letter>/<package_name>/<package_file>',
           methods=['GET', 'HEAD'])
def package(package_type, letter, package_name, package_file):
    ''' Downloads the egg

    :param str package_type: the nature of the package. For example:
                              'source' or '2.7'
    :param str letter: the first char of the package name. For example:
                              D
    :param str package_name: the name of the package. For example: Django
    :param str package_file: the name of the package and it's version. For
                             example: Django-1.5.0.tar.gz
    '''
    egg_filename = join(get_base_path(), package_name, package_file)

    remote = request.args.get('remote')
    if remote:
        # the requested link is not on pypi.python.org, we need to use the remote URL
        url = remote
    else:
        url = app.config['PYPI_URL'] + 'packages/%s/%s/%s/%s' % (package_type,
                                                                 letter,
                                                                 package_name,
                                                                 package_file)
    if request.method == 'HEAD':
        # in this case the content type of the file is what is
        # required
        if not exists(egg_filename):
            pypi_response = head(url)
            return _respond(pypi_response.content, pypi_response.headers['content-type'])

        else:
            mimetype = magic.from_file(egg_filename, mime=True)
            return _respond('', mimetype)

    app.logger.debug('Downloading: %s', package_file)
    if exists(egg_filename):
        app.logger.debug('Found local file in repository for: %s', package_file)
        # if the file exists, then use the local file.
        path = get_package_path(package_name)
        path = join(path, package_file)
        with open(path, 'rb') as egg:
            content = egg.read(-1)
        mimetype = magic.from_file(egg_filename, mime=True)
        return _respond(content, mimetype)

    else:
        # Downloads the egg from pypi and saves it locally, then
        # it will return it.
        package_path = get_package_path(package_name)
        app.logger.debug('Starting to download: %s using the url: %s',
                         package_file, url)
        pypi_response = get(url)
        app.logger.debug('Finished downloading package: %s', package_file)

        if pypi_response.status_code != 200:
            app.logger.warning('Error respose while downloading for proxy: %s'
                               'Response details: %s', package_file,
                               pypi_response.text)
            abort(pypi_response.status_code)

        if not exists(package_path):
            makedirs(package_path)

        with open(egg_filename, 'w') as egg_file:
            egg_file.write(pypi_response.content)

        with open(egg_filename) as egg_file:
            filecontent = egg_file.read(-1)
            mimetype = magic.from_file(egg_filename, mime=True)

        with open(egg_filename + '.md5', 'w') as md5_output:
            md5 = get_md5_for_content(filecontent)
            md5_output.write(md5)

        return _respond(filecontent, mimetype)


def _respond(filecontent, mimetype):
    return make_response(filecontent, 200, {
                        'Content-Type': mimetype
                    }
            )



# -*- coding: utf-8 -*-

''' Handles the submition of the different packages.

This is used on **register** o **upload**.

For example:

.. code-block:: bash

    python setup.py register
    python setup.py sdist upload


'''

from os import makedirs
from os.path import exists, join

from flask import request
from werkzeug import secure_filename

from flask_pypi_proxy.utils import get_package_path
from flask_pypi_proxy.app import app


@app.route('/pypi/', methods=['POST'])
def index():
    path = get_package_path(request.form['name'])
    if not exists(path):
        makedirs(path)

    if request.files:
        file = request.files['content']
        filename = secure_filename(file.filename)
        file.save(join(path, filename))
        with open(join(path, filename + '.md5'), 'w') as md5_file:
            md5_file.write(request.form['md5_digest'])
    return 'Registered'



#!/bin/env python

from setuptools import setup, find_packages

setup(
    name='Flask-Pypi-Proxy',
    version='0.4.2',
    description='A Pypi proxy',
    long_description=open('README.rst').read(-1),
    author='Tomas Zulberti',
    author_email='tzulberti@gmail.com',
    license='BSD',
    url='https://github.com/tzulberti/Flask-PyPi-Proxy',
    install_requires=[
        "Flask",
        "requests",
        "python-magic",
        "beautifulsoup4"
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    keywords='pypi flask proxy',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



# -*- coding: utf-8 -*-

from flask_pypi_proxy.app import app
import flask_pypi_proxy.views.package
import flask_pypi_proxy.views.pypi
import flask_pypi_proxy.views.simple


if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0')



# -*- coding: utf-8 -*-

''' The main Flask application which take care of reading the configuration.
'''

import os
import json
import logging
from flask import Flask

app = Flask(__name__)


def read_configuration(app, pypi_url='http://pypi.python.org',
                        private_eggs=[]):
    ''' Reads the configuration by using the system file or the configuration
    file.

    :param :class:`Flask` app: the app to where set the configuration values

    :param str pypi_url: the base Pypi url from where get the packages.

    :param str private_eggs: the list of the name of the proyects for which
                             the **pypi_url** won't be used.
    '''
    filepath = os.environ.get('FLASK_PYPI_PROXY_CONFIG')
    if filepath:
        # then the configuration file is used
        if not os.path.exists(filepath):
            raise Exception('There should be a configuration file but '
                            'the path is invalid')

        with open(filepath) as config_file:
            configuration = json.load(config_file)
            if not 'BASE_FOLDER_PATH' in configuration:
                raise Exception('The configuration should have the value '
                                'for BASE_FOLDER_PATH')

            if not 'LOGGING_PATH' in configuration:
                raise Exception('The configuration must have the value for '
                                'LOGGING_PATH')

            app.config['PRIVATE_EGGS'] = configuration.get('PRIVATE_EGGS',
                                                           private_eggs)
            app.config['BASE_FOLDER_PATH'] = configuration['BASE_FOLDER_PATH']
            app.config['SHOULD_USE_EXISTING'] = configuration.get(
                                                        'SHOULD_USE_EXISTING',
                                                        False)
            app.config['PYPI_URL'] = configuration.get('PYPI_URL', pypi_url)
            app.config['LOGGING_PATH'] = configuration.get('LOGGING_PATH')
            app.config['LOGGING_LEVEL'] = configuration.get('LOGGING_LEVEL',
                                                            'DEBUG')

    else:
        base_path = os.environ.get('PYPI_PROXY_BASE_FOLDER_PATH')
        logging_path = os.environ.get('PYPI_PROXY_LOGGING_PATH')
        if not base_path:
            raise Exception('The PYPI_PROXY_BASE_FOLDER_PATH or '
                            'FLASK_PYPI_PROXY_CONFIG environment keys '
                            'are missing')

        if not logging_path:
            raise Exception('The PYPI_PROXY_LOGGING_PATH or '
                            'FLASK_PYPI_PROXY_CONFIG environment keys '
                            'are missing')

        app.config['BASE_FOLDER_PATH'] = base_path
        app.config['LOGGING_PATH'] = logging_path
        app.config['LOGGING_LEVEL'] = os.environ.get('PYPI_PROXY_LOGGING_LEVEL',
                                                     'DEBUG')
        env_value = os.environ.get('PYPI_PROXY_PRIVATE_EGGS', None)
        if env_value:
            app.config['PRIVATE_EGGS'] = env_value.split(',')
        else:
            app.config['PRIVATE_EGGS'] = private_eggs

        app.config['PYPI_URL'] = os.environ.get('PYPI_PROXY_PYPI_URL',
                                                pypi_url)
        app.config['SHOULD_USE_EXISTING'] = os.environ.get(
                                            'PYPI_PROXY_SHOULD_USE_EXISTING',
                                            False)

    if not app.config['PYPI_URL'].endswith('/'):
        app.config['PYPI_URL'] = app.config['PYPI_URL'] + '/'


def configure_logging(app):
    ''' Setups the logging that will be used by the views to log the
    different problems that it might have.

    :param app: the application that will be used to register the URLs.
    '''
    if not app.debug:
        logging.basicConfig(filename=app.config['LOGGING_PATH'],
                            level=getattr(logging, app.config['LOGGING_LEVEL']),
                            format='%(asctime)s [%(levelname)s] %(message)s')

read_configuration(app)
configure_logging(app)
